
# source('DES/utils.R')

perform_DES = function(horizon , department_names,  patienttype_info, current_patient_info, return_logbook = FALSE, start_date = NULL, by_day = T){
  # this function performs discrete event simulation for a certain horizon, patienttype_info is a list containing information about the type
  # of patients in the system
  cur_time = 0                 # current time in simulation (should be less than horizon)                                 
  
  # for each patient-type in the system, determine the next arrival times from the patienttype_info
  arrival_times_per_type = c() 
  for(i in seq(length(patienttype_info))){
    ptype_info = patienttype_info[[i]]
    arrival_distribution = ptype_info$arrival_distribution
    if(is.null(arrival_distribution) ){
      arrival_times_per_type[names(patienttype_info)[i]] = Inf
    }else{
      if(strcmp(arrival_distribution, 'NHPois')){
        arrival_times_per_type[names(patienttype_info)[i]] =  rnhpois(cur_time, ptype_info$arrival_info$cuml_intensity_NHPP,
                                                                      ptype_info$arrival_info$icuml_intensity_NHPP)
      }else{
        if(strcmp(arrival_distribution, 'Scheduled')){
          arrival_schedule                                   = ptype_info$arrival_schedule
          res                                                = get_next_scheduled_arrival(cur_time, arrival_schedule)
          arrival_times_per_type[names(patienttype_info)[i]] = res$arrivaltime
          patienttype_info[[i]]$arrival_schedule = res$schedule
        }}}}
  
  departure_times = c(Inf)                             # list containing departure times inf is needed to stop the simulation from breaking in an empty system
  state           = zeros(length(department_names),1)  # system state
  names(state)    = department_names
  
  #determining the departure times and LOS of the patients currently in the system, given in current_patient_info
  list_pat_in_system =   list()# list of the patients currently in the system
  for(current_patient in current_patient_info){
    #determining type of patient
    current_patient$type = names(which(cumsum(current_patient$prob_types)>=runif(1))[1])
    
    #if unknown, determining departure info from patient type
    if(is.null(current_patient$departure_info)){
      current_patient$departure_info = (patienttype_info[[current_patient$type]])$departure_dist_info
    }
    
    #sampling remaining length of stay
    remaining_LOS_pat                      = sample_remaining_LOS(current_patient$LOS_init, current_patient$departure_info$cdf, 
                                                                  current_patient$departure_info$icdf) 
    
    #bookkeeping, storing the departure times and LOS for the patient, storing departure time in departure_times and incrementing the initial state
    current_patient$departure_time         = remaining_LOS_pat
    departure_times                        = c(departure_times, remaining_LOS_pat )
    current_patient$LOS                    = current_patient$LOS_init + remaining_LOS_pat
    state[current_patient$department]      = state[current_patient$department]  + 1
    list_pat_in_system                     = append(list_pat_in_system, list(current_patient))
  }
  
  
  trajectory = state # trajectory will be a time series of the total number of patients at the departments
  jump_times = c(0)  # jump times will be a sequence of times at which the state changes
  logbook = list()   
  while(cur_time <= horizon){
    min_T_arrival   = min(arrival_times_per_type)
    min_T_departure = min(departure_times)
    
    if(is.infinite(min_T_arrival) & is.infinite(min_T_departure)){
      break
    }
    if(min_T_arrival < min_T_departure){ # if the next event is an arrival
      
      cur_time           = min_T_arrival # set the current time to the arrival time
      
      #determine the type of the arriving patient and 
      arriving_ptype = names(which(arrival_times_per_type == min_T_arrival)[1])
      new_patient    = patienttype_info[[arriving_ptype]]
      
      
      # fill in the patient-linked details
      new_patient$type           = arriving_ptype
      
      #sampling/storing length of stay (and hence departure time) of arriving patient
      LOS_newpat                 = new_patient$departure_dist_info$icdf(runif(1))
      new_patient$LOS            = LOS_newpat
      new_departure_time         = cur_time + LOS_newpat
      departure_times            = c(departure_times, new_departure_time)
      new_patient$departure_time = new_departure_time
      
      # sampling/storing department at which the patient arrives
      department_arrival         = names(first(which(cumsum(new_patient$department_arrival_info)>=runif(1))))
      state[department_arrival]  = state[department_arrival] + 1
      new_patient$department     = department_arrival
      
      # adding patients to patient list
      list_pat_in_system         = append(list_pat_in_system,list(new_patient)  )
      
      # sampling new arrival time for the patient type (for next arrival)
      arrival_distribution = new_patient$arrival_distribution
      if(strcmp(arrival_distribution, 'NHPois')){
        arrival_times_per_type[arriving_ptype] =  rnhpois(cur_time, new_patient$arrival_info$cuml_intensity_NHPP,
                                                          new_patient$arrival_info$icuml_intensity_NHPP)
      }else{
        if(strcmp(arrival_distribution, 'Scheduled')){
          arrival_schedule                                    = new_patient$arrival_schedule
          res                                                 = get_next_scheduled_arrival(cur_time, arrival_schedule)
          arrival_times_per_type[arriving_ptype]              = res$arrivaltime
          patienttype_info[[arriving_ptype]]$arrival_schedule = res$schedule
        }}
      
    }else{ # if the next event is a departure
      
      cur_time   = min_T_departure  # set the current time to the departure time
      
      #finding the (index of the) departing patient
      idx_departing_pat                         = which(departure_times == min_T_departure)[[1]]
      departing_patient                         = list_pat_in_system[[idx_departing_pat-1]]
      
      # decrementing the state at the corresponding department
      state[departing_patient$department]       = state[departing_patient$department] - 1
      
      # removing the corresponding patient from the system (in case of transition the patient is added later again)
      departure_times                           = departure_times[-idx_departing_pat]
      list_pat_in_system[[idx_departing_pat-1]] = NULL
      
      #determining the type of the departing patient
      info_ptype                                = patienttype_info[[departing_patient$type]]
      
      #finding transition probabilities for the patient to go to another department given his/her LOS
      LOS_departing_pat      = departing_patient$LOS
      if(is.null(info_ptype$departure_transition_info)){
        trans_probs_department = 0
      }else{
        trans_probs_department = sapply(info_ptype$departure_transition_info, function(f){f(LOS_departing_pat)})
      }
      
      uniform_draw           = runif(1)
      if(uniform_draw < sum(trans_probs_department)){ # if the patient does not go out of the system..
        
        #find the department the patient goes to, increment state there
        new_department          = names(first(which(cumsum(trans_probs_department)>uniform_draw)))
        state[[new_department]] = state[[new_department]] + 1
        
        # find the new patient-type for this patient
        trans_probs_type = sapply(info_ptype$type_transition_info, function(f){f(LOS_departing_pat)})
        new_typeof_pat   = names(first(which(cumsum(trans_probs_type) > runif(1))))
        
        # determine the list of patient details necessary for this "new patient" and store
        new_patient      = patienttype_info[[new_typeof_pat]]
        new_patient$type           = new_typeof_pat
        LOS_newpat                 = new_patient$departure_dist_info$icdf(runif(1))
        new_patient$LOS            = LOS_newpat
        new_departure_time         = cur_time + LOS_newpat 
        departure_times            = c(departure_times, new_departure_time )
        new_patient$departure_time = new_departure_time
        new_patient$department     = new_department
        
        list_pat_in_system         = append(list_pat_in_system, list(new_patient))
      }}
    
    #updating trajectory
    trajectory        = rbind(trajectory, state) #update trajectory
    jump_times        = c(jump_times, cur_time)
    
    #updating logbook (for debugging puposes)
    if(return_logbook){
      page = list(cur_time,  arrival_times_per_type, sapply(list_pat_in_system, function(L){L$LOS}), 
                  sapply(list_pat_in_system, function(L){L$type}), 
                  sapply(list_pat_in_system, function(L){L$departure_time}),
                  sapply(list_pat_in_system, function(L){L$department}), state)
      names(page) = c('time', 'arrival times', 'LOS cur pat' , 'types cur pat', 'dep times cur pat', 'dept cur pat', 'state system')
      
      logbook = append(logbook, list(page))
    }
    
  }
  
  if(is.null(start_date)){
    rownames(trajectory) = jump_times
  }else{# assumes the jump times are in days when the start date of the simulation (entry in time_info) is given
    dates = start_date + ddays(jump_times)
    trajectory = add_column(as_tibble(trajectory), date = dates, .before = 1)
  }
  
  if(by_day){
    days_pred = seq(start_date, start_date+days(horizon), by = 'days')
    days_traj = tibble(date = days_pred)
    
    #definieer de bezetting op een dag als de bezetting aan het begin van die dag (let op ceiling date)
    trajectory = trajectory %>% mutate(date = ceiling_date(date, unit = 'days',change_on_boundary = F)) %>% group_by(date) %>% summarise(Kliniek = last(Kliniek), IC = last(IC), .groups = 'drop') %>% ungroup() 
    trajectory = days_traj %>% left_join(trajectory, by = 'date', keep =F) %>% fill(c(IC,Kliniek), .direction = 'down') %>% fill(c(IC,Kliniek), .direction = 'up')
  }
  
  if(return_logbook){
    return(list(trajectory, logbook))
  }else{
    return(trajectory)}}


if(sys.nframe() == 0){
  
  # determining departure information, transition function and arrival intensity information for poisson process
  departure_std_info = list(cdf = function(t){1-exp(-t)}, icdf = function(t)(-log(1-t))) #exponential service times
  transition_fn = function(t){0.2} #0.2 constant
  arrival_intensity_std = list(cuml_intensity_NHPP = function(t){2*t}, icuml_intensity_NHPP = function(t){t/2}) #  rate 2 homogeneous poisson
  
  #setting up patient type information lists, containing for each patient type:
  # - the type of arrival distribution, and the information for this process
  # - the type of departure distribution for the patient type
  # - transition functions that determine (based on LOS) the probability that the patient goes to another department after leaving the current one
  # - transition functions that determine (based on length of stay) the probability that the patient becomes another patient type after completing stay at department
  
  patienttype_info = list(type_1 = list(arrival_info              = arrival_intensity_std,
                                        departure_dist_info       = departure_std_info,
                                        departure_transition_info = c(Kliniek = constantfn(0), IC = transition_fn),
                                        type_transition_info      = c(type_1 = constantfn(0), type_2 = constantfn(1)),
                                        department_arrival_info   = c(Kliniek = 1, IC = 0),
                                        arrival_distribution      = 'NHPois'),
                          
                          type_2 = list(arrival_info              = arrival_intensity_std,
                                        departure_dist_info       = departure_std_info,
                                        departure_transition_info = c(Kliniek = transition_fn, IC = constantfn(0)),
                                        type_transition_info      = c(type_1 = constantfn(1), type_2 = constantfn(0)),
                                        department_arrival_info   = c(Kliniek = 0, IC = 1),
                                        arrival_distribution      = 'NHPois'))
  
  
  #current LOS, departments, type, departure cdf's and icdf's  for all patients currently in the system (can be patient-specific)
  current_patient_info = append(lapply(c(9,3,3,6,2,18,30,4,7,4), function(LOS){return(list(LOS_init = LOS,
                                                                                           department = 'Kliniek',
                                                                                           prob_types = list(type_1 = 1 , type_2 = 0)))}),
                                lapply(c(2,5,20,5,1,1,8,1,10,12), function(LOS){return(list(LOS_init = LOS,
                                                                                            department = 'IC',
                                                                                            prob_types = list(type_1 = 0 , type_2 = 1)))}))
  
  department_names = c('Kliniek', 'IC') # names for all departments in the system
  
  #determining simulation horizon
  horizon = 30 
  
  #performing simulations
  M = 1000 #num sim
  simres = list()
  prediction = zeros(horizon+1, 2)
  for(i in seq(M)){
    res  = perform_DES(horizon , department_names , patienttype_info, current_patient_info)
    jump_times = as.numeric(rownames(res))
    
    #resample the output to state evaluation per day.
    timepoints_intpl = seq(0, horizon)
    res = apply(res, 2, function(v){ approxfun(jump_times, v)(timepoints_intpl)})
    prediction = prediction + res/M
  }
  
  # obtaining average number of patients per time step
  rownames(prediction) = timepoints_intpl
  colnames(prediction) = sapply(seq(dim(res)[[2]]), function(i){paste('dept ', toString(i), sep = '')})
  
  #plot
  plot(rownames(prediction), prediction[,1], type = 'l')   #long-term LOS is 2.5 due to Little's Law
  lines(rownames(prediction), 2.5*ones(length(rownames(prediction)),1), lty = 2, col = 'red', lwd = 2)
  plot(rownames(prediction), prediction[,2], type = 'l')   #long-term LOS is 2.5 due to Little's Law
  lines(rownames(prediction), 2.5*ones(length(rownames(prediction)),1), lty = 2, col = 'red', lwd = 2)
}