# Init ----------------------------------------------------------------------------------------



# Packages
suppressPackageStartupMessages(library(tidyverse, quietly = TRUE))
library(lubridate, warn.conflicts = FALSE)
library(investr)
library(nlme, warn.conflicts = FALSE)
library(FlexParamCurve)
library(readxl)
library(survival, warn.conflicts = FALSE)
library(magrittr, warn.conflicts = FALSE)
library(ggpubr)
library(survminer, warn.conflicts = FALSE)
library(readr)

# Instellingen voor locale
options(readr.num_columns = 0)
nl <- locale(date_names = "nl", decimal_mark = ",", grouping_mark = ";")
options(readr.default_locale = nl)

# Instellingen voor grafieken
theme_set(theme_minimal(base_size = 15))

# Functies ------------------------------------------------------------------------------------

laad_nieuwe_opnames <- function(ziekenhuis, prefix = "") {
  map <- paste0("data/", ziekenhuis, "_nieuwe_opnames")

  if (!prefix == "") {
    pattern <- paste0("^", prefix)
  } else {
    pattern <- "^2020.*csv$"
  }

  # Zoek het laatste bestand (op alfabetische volgorde)
  bestand <- tail(list.files(map, pattern, full.names = TRUE), n = 1)

  df <- read_csv2(bestand) %>%
    mutate(num = 1:nrow(.))

  df$weekend <- ifelse(weekdays(df$datum) %in% c("zaterdag", "zondag"), 1, 0)

  print(paste("Bestand geladen:", bestand))

  return(df)
}


laad_opnames <- function(ziekenhuis = "etz", prefix = "") {
  map <- paste0("data/", ziekenhuis, "_opnames")
  pattern <- paste0("^", prefix, ".*csv$")

  # Zoek het laatste bestand (op alfabetische volgorde)
  bestand <- tail(list.files(map, pattern, full.names = TRUE), n = 1)

  df <- read_csv2(bestand)
  print(paste("Bestand geladen:", bestand))
  return(df)
}


laad_output <- function(naam, ziekenhuis = "etz", prefix = "") {
  map <- paste0("output/", naam)

  if (prefix == "") {
    pattern <- paste0("^", ziekenhuis, "_2020")
  } else {
    pattern <- paste0("^", ziekenhuis, "_", prefix, "_2020")
  }

  # Zoek het laatste bestand (op alfabetische volgorde)
  bestand <- tail(list.files(map, pattern, full.names = TRUE), n = 1)

  out <- read_delim(bestand, locale = nl, delim = ";", trim_ws = TRUE)

  print(paste("Bestand geladen:", bestand))

  return(out)
}


voorspel_nieuwe_opnames <- function(preds) {
  out <- preds %>%
    mutate(
      nieuwe_opn = obs - lag(obs, default = 0),
      log = fit - lag(fit, default = NA),
      log_lwr = lwr - lag(fit, default = NA),
      log_upr = upr - lag(fit, default = NA),
      # De schatting kan niet lager dan 0 zijn
      log = ifelse(log < 0, 0, log),
      log_lwr = ifelse(log_lwr < 0, 0, log_lwr),
      log_upr = ifelse(log_upr < 0, 0, log_upr)
    ) %>%
    mutate_if(is_double, ~ round(., 1))

  return(out)
}

schrijf_rds <- function(object, ziekenhuis) {
  bestand <- paste0("rds/", ziekenhuis, "_", deparse(substitute(object)), ".rds")

  write_rds(object, bestand)
}


laad_rds <- function(naam, ziekenhuis = "etz") {
  if (basename(getwd()) == "rmarkdown") {
    map <- "../rds/"
  } else {
    map <- "rds/"
  }

  bestand <- paste0(map, ziekenhuis, "_", naam, ".rds")

  read_rds(bestand)
}

bepaal_ymax <- function(df, round = 10) {
  ymax <- df %>%
    select_if(is.numeric) %>%
    max(na.rm = T)

  ymax <- round * ceiling(ymax / round)

  return(ymax)
}

plot_weekend <- function(df, round_ymax = 10) {
  ymax <- bepaal_ymax(df, round_ymax)

  df <- tibble(
    datum = df$datum,
    weekend = ifelse(weekdays(datum) %in% c("zaterdag", "zondag"), ymax, 0)
  )

  return(
    geom_bar(aes(datum, weekend),
      data = df, stat = "identity", width = 1,
      alpha = .2
    )
  )
}

add_title <- function(title, ziekenhuis) {
  ggtitle(
    paste0(title, " - ", toupper(ziekenhuis))
  )
}

# Als de maximale ligduur groter is dan max. conditional probabilities, voeg kolommen
# toe
voeg_kolommen_condprobs_toe <- function(condprobs, opgenomen) {
  if (ncol(condprobs) < (max(opgenomen$ligduur) + 1)) {
    nieuwe_kol_nums <- ncol(condprobs):max(opgenomen$ligduur)
    nieuwe_kol_namen <- paste0("ligduur_", str_pad(nieuwe_kol_nums, 2, pad = "0"))

    # Aanname: deze patiënten gaan allemaal weg (want liggen al heel lang)
    condprobs[nieuwe_kol_namen] <- 0
  }

  return(condprobs)
}

# Maak een data frame met als kolom ligduur en als waarde het aantal patiënten met die
# ligduur. Als condprobs meer kolommen heeft, voeg kolommen met waarde 0 toe.
bereken_pat_per_ld <- function(opgenomen, condprobs) {
  pat_per_ld <- tibble(
    # Maak evenveel kolommen als in condprobs
    ligduur = 1:(ncol(condprobs) - 1)
  ) %>%
    left_join(opgenomen, by = "ligduur") %>%
    replace_na(list(aantal = 0)) %>%
    select(ligduur, aantal) %>%
    mutate(ligduur = paste0("ligduur_", str_pad(ligduur, 2, pad = "0"))) %>%
    spread(ligduur, aantal) %>%
    select(everything())

  return(pat_per_ld)
}

voorspel_doorstroom_toekomst <- function(preds, condprobs, aantal_dagen) {

  # Elke rij 21 keer kopiëren
  # Condprobs 7 keer kopiëren
  # Elke datum plus de 6 dagen daarna
  datums <- map_dfr(
    preds$datum, ~ data.frame(datum = .x + days(0:(aantal_dagen - 1)))
  ) %>%
    pull(datum)
  # Herhaal de kansen
  prob_doorstroom <- rep(condprobs$ligduur_01, aantal_dagen)
  
  doorstroom_toekomst <- preds %>% 
    select(-datum) %>% 
    slice(rep(1:nrow(.), each = aantal_dagen)) %>% 
    mutate_all(~ .x * prob_doorstroom) %>% 
    mutate(datum = datums) %>% 
    group_by(datum) %>% 
    summarize_all(sum) %>% 
    slice(1:aantal_dagen)

  # omzetten naar instroom ipv cumulatief
  doorstroom <- doorstroom_toekomst %>%
    mutate_at(
      vars(-datum),
      ~ .x - lag(.x, default = 0)
    )

  return(doorstroom)
}

voorspel_bezetting_toekomst <- function(preds, condprobs, aantal_dagen) {
  # Voor alle toekomstige patiënten, wanneer stromen ze uit?

  # Elke rij 21 keer kopiëren
  # Condprobs 7 keer kopiëren
  # Elke datum plus de 6 dagen daarna
  datums <- map_dfr(
    preds$datum, ~ data.frame(datum = .x + days(0:(aantal_dagen - 1)))
  ) %>%
    pull(datum)
  # Herhaal de kansen
  prob_survival <- rep(condprobs$ligduur_01, aantal_dagen)
  
  preds %>% 
    select(-datum) %>% 
    slice(rep(1:nrow(.), each = aantal_dagen)) %>% 
    mutate_all(~ .x * prob_survival) %>% 
    mutate(datum = datums) %>% 
    group_by(datum) %>% 
    summarize_all(sum) %>% 
    slice(1:aantal_dagen)
}
