est_condprobs <- function(fit, max_ligduur = 30) {

  # Aanname: Na 30 dagen is iedereen weg. Deze aanname wordt niet gebruikt als er in de
  # dataset al patiënten zitten die er langer dan 30 dagen liggen.
  # TODO: Dit is voor IC misschien te optimistisch?
  max_ligduur <- max(max(fit$time), max_ligduur)

  df_surv_gevuld <- get_surv_params(fit, max_ligduur)

  # Maak een data frame met conditional probabilities:
  # - Kolommen: Aantal dagen vooruit kijken (Bijv. condprob_01 is de kans dat je er
  #   morgen nog ligt)
  # - Rijen: Aantal dagen dat je er al ligt (Bijv. rij 1 is dat je er 1 dag ligt, dus
  # één dag na aankomst. Hier voegen we later nog de onvoorwaardelijke kansen aan toe.)

  condprobs_dagen_vooruit <- bereken_condprobs_dagen_vooruit(
    df_surv_gevuld$surv, max_ligduur
  )

  # Transformeer data frame door kolommen en rijen om te wisselen:
  # - Kolommen: Aantal dagen dat je er al ligt (bijv. ligduur_01 is patiënten die
  # vandaag zijn aangekomen)
  # - Rijen: Aantal dagen vooruit kijken (bijv. aantal_dagen_vooruit = 1 is de kans dat
  # ze vandaag weggaan)

  # Voeg een rij met nullen toe zodat iedereen uiteindelijk weggaat
  df_surv_gevuld <- df_surv_gevuld %>%
    bind_rows(
      tibble(time = max_ligduur + 1, surv = 0, std.err = 0, lower = 0, upper = 0)
    )

  condprobs <- transpose_condprobs(
    condprobs_dagen_vooruit, df_surv_gevuld$surv
  )

  # Bereken lower en upper bound m.b.v. de standard error voor dat aantal dagen vooruit
  condprobs_lwr <- confint_condprobs(condprobs, "lwr", df_surv_gevuld)
  condprobs_upr <- confint_condprobs(condprobs, "upr", df_surv_gevuld)

  return(list(est = condprobs, lwr = condprobs_lwr, upr = condprobs_upr))
}


transpose_condprobs <- function(df, surv_probs) {
   df %>%
    pivot_longer(-time, names_to = "aantal_dagen_vooruit") %>%
    mutate(time = paste0("ligduur_", str_pad(time, 2, pad = "0"))) %>%
    pivot_wider(names_from = time, values_from = value) %>%
    mutate(ligduur_01 = surv_probs) %>%
    select(aantal_dagen_vooruit, ligduur_01, everything()) %>%
    mutate(aantal_dagen_vooruit = 1:nrow(.))
  
}


confint_condprobs <- function(condprobs, direction, df_surv) {
  if (direction == "lwr") {
    z_value <- -1.96
    unconditional_bounds <- df_surv$lower
  } else {
    z_value <- 1.96
    unconditional_bounds <- df_surv$upper
  }

  condprobs %>%
    mutate_at(vars(-aantal_dagen_vooruit), ~ .x + z_value * df_surv$std.err) %>%
    mutate(ligduur_01 = unconditional_bounds)
}


get_surv_params <- function(fit, max_ligduur, perc = NULL) {
  # Maak een data frame met de survival parameters
  df_surv <- tibble(
    time = fit$time,
    surv = fit$surv,
    std.err = fit$std.err * fit$surv,
    lower = fit$lower,
    upper = fit$upper
  )

  df_surv_gevuld <- tibble(time = 1:max_ligduur) %>%
    left_join(df_surv, by = "time") %>%
    fill(everything(), .direction = "down") %>%
    replace_na(list(
      surv = 1,
      std.err = 0,
      lower = 1,
      upper = 1
    ))

  if (!is.null(perc)) {
    df_surv_gevuld <- transformeer_surv_params(df_surv_gevuld, perc)
  }

  return(df_surv_gevuld)
}


transformeer_surv_params <- function(df, perc) {
  mutate_at(df, vars(surv, lower, upper), ~ perc - perc * .x)
}


bereken_condprobs_dagen_vooruit <- function(surv_probs, max_ligduur) {
  out <-
    map_dfc(
      maak_namen_condprobs(max_ligduur),
      ~ bereken_condprobs(surv_probs, .x)
    )

  out$time <- 2:(max_ligduur + 1)

  drop_na(out)
}


maak_namen_condprobs <- function(max_ligduur) {
  ligduren <- 1:(max_ligduur + 1)
  var_namen <- paste0("condprob_", str_pad(ligduren, 2, pad = "0"))
  set_names(ligduren, var_namen)
}


bereken_condprobs <- function(surv, ligduur) {
  lead(surv, n = ligduur, default = 0) / surv
}


# Functies doorstroom -------------------------------------------------------------

est_condprobs_doorstroom <- function(fit, fit_tot, perc, max_ligduur = 30) {

  # Aanname: Na 30 dagen is iedereen weg. Deze aanname wordt niet gebruikt als er in de
  # dataset al patiënten zitten die er langer dan 30 dagen liggen.
  # TODO: Dit is voor IC misschien te optimistisch?
  max_ligduur <- max(max(fit$time), max_ligduur)

  df_surv_gevuld <- get_surv_params(fit, max_ligduur, perc)

  # Maak een data frame met conditional probabilities:
  # - Kolommen: Aantal dagen vooruit kijken (Bijv. condprob_01 is de kans dat je er
  #   morgen nog ligt)
  # - Rijen: Aantal dagen dat je er al ligt (Bijv. rij 1 is dat je er 1 dag ligt, dus
  # één dag na aankomst. Hier voegen we later nog de onvoorwaardelijke kansen aan toe.)

  condprobs_dagen_vooruit <- bereken_condprobs_dagen_vooruit_doorstroom(
    df_surv_gevuld$surv, max_ligduur, perc
  )

  # Transformeer data frame door kolommen en rijen om te wisselen:
  # - Kolommen: Aantal dagen dat je er al ligt (bijv. ligduur_01 is patiënten die
  # vandaag zijn aangekomen)
  # - Rijen: Aantal dagen vooruit kijken (bijv. aantal_dagen_vooruit = 1 is de kans dat
  # ze vandaag weggaan)

  # Herhaal de laatste rij
  df_surv_gevuld <- df_surv_gevuld %>%
    bind_rows(
      slice(df_surv_gevuld, nrow(df_surv_gevuld))
    )

  condprobs <- transpose_condprobs(
    condprobs_dagen_vooruit, df_surv_gevuld$surv
  )
  
  #toegevoegd 2021-01-11
  df_surv_tot = get_surv_params(fit_tot, max_ligduur)
  for(i in seq(1,ncol(condprobs)-2)){
    
    condprobs_trans = (condprobs$ligduur_01[-seq(i)] - condprobs$ligduur_01[i])/df_surv_tot$surv[i]
    condprobs[,i+2] = c(condprobs_trans, rep(tail(condprobs_trans,1), nrow(condprobs) - length(condprobs_trans)))
  }

  # Bereken lower en upper bound m.b.v. de standard error voor dat aantal dagen vooruit
  condprobs_lwr <- confint_condprobs(condprobs, "lwr", df_surv_gevuld)
  condprobs_upr <- confint_condprobs(condprobs, "upr", df_surv_gevuld)

  return(list(est = condprobs, lwr = condprobs_lwr, upr = condprobs_upr))
}


bereken_condprobs_dagen_vooruit_doorstroom <- function(surv_probs, max_ligduur, perc) {
  prop_over <- 1 - surv_probs

  out <- map_dfc(
    maak_namen_condprobs(max_ligduur),
    ~ 1 - lead(prop_over, n = .x, default = 1 - perc) / prop_over
  )

  out$time <- 2:(max_ligduur + 1)

  drop_na(out)
}


bereken_condprobs_doorstroom <- function(surv, ligduur, perc) {
  # Variabele voor proportie die nog niet vertrokken is
  prop_over <- 1 - surv

  # Bereken de conditionele kans dat iemand blijft, en trek die af van 1
  1 -
    lead(prop_over, n = ligduur, default = 1 - perc) / prop_over
}
